package server

//go:generate python sri.py templates/*.html
//go:generate go-bindata --nocompress --pkg server static/... templates/...

import (
	"encoding/gob"
	"fmt"
	"html/template"
	"io"
	"log"
	"net/http"
	"net/url"
	"strings"
	"time"

	assetfs "github.com/elazarl/go-bindata-assetfs"
	"github.com/gorilla/csrf"
	"github.com/gorilla/mux"
	"github.com/gorilla/sessions"

	"git.autistici.org/id/auth"
	authclient "git.autistici.org/id/auth/client"
	"git.autistici.org/id/go-sso/httputil"
	"git.autistici.org/id/go-sso/server/device"
)

const authSessionKey = "_auth"

type authSession struct {
	*httputil.ExpiringSession

	// User name and other information (like group membership).
	Username string
	UserInfo *auth.UserInfo

	// Services the user has logged in to from this session.
	Services []string
}

// AddService adds a service to the current session (if it's not
// already there).
func (s *authSession) AddService(service string) {
	for _, svc := range s.Services {
		if svc == service {
			return
		}
	}
	s.Services = append(s.Services, service)
}

// By default, make users log in again after (almost) one day.
var defaultAuthSessionLifetime = 20 * time.Hour

func newAuthSession(ttl time.Duration, username string, userinfo *auth.UserInfo) *authSession {
	return &authSession{
		ExpiringSession: httputil.NewExpiringSession(ttl),
		Username:        username,
		UserInfo:        userinfo,
	}
}

func init() {
	gob.Register(&authSession{})
}

// Returns the URL of the login handler on the target service.
func serviceLoginCallback(service, destination, token string) string {
	v := make(url.Values)
	v.Set("t", token)
	v.Set("d", destination)
	return fmt.Sprintf("https://%ssso_login?%s", service, v.Encode())
}

// Returns the URL of the logout handler on the target service.
func serviceLogoutCallback(service string) string {
	return fmt.Sprintf("https://%ssso_logout", service)
}

// Server for the SSO protocol. Provides the HTTP interface to a
// LoginService.
type Server struct {
	authSessionStore    sessions.Store
	authSessionLifetime time.Duration
	loginHandler        *loginHandler
	loginService        *LoginService
	csrfSecret          []byte
	tpl                 *template.Template
}

func sl2bl(sl []string) [][]byte {
	var out [][]byte
	for _, s := range sl {
		out = append(out, []byte(s))
	}
	return out
}

// New returns a new Server.
func New(loginService *LoginService, authClient authclient.Client, config *Config) (*Server, error) {
	sessionSecrets := sl2bl(config.SessionSecrets)
	store := sessions.NewCookieStore(sessionSecrets...)
	store.Options = &sessions.Options{
		HttpOnly: true,
		Secure:   true,
		MaxAge:   0,
		Path:     "/",
	}
	s := &Server{
		authSessionLifetime: defaultAuthSessionLifetime,
		authSessionStore:    store,
		loginService:        loginService,
		tpl:                 parseEmbeddedTemplates(),
	}
	if config.CSRFSecret != "" {
		s.csrfSecret = []byte(config.CSRFSecret)
	}
	if config.AuthSessionLifetimeSeconds > 0 {
		s.authSessionLifetime = time.Duration(config.AuthSessionLifetimeSeconds) * time.Second
	}

	devMgr, err := device.New(config.DeviceManager)
	if err != nil {
		return nil, err
	}
	s.loginHandler = newLoginHandler(s.loginCallback, devMgr, authClient, config.AuthService, s.tpl, sessionSecrets...)

	return s, nil
}

func (h *Server) loginCallback(w http.ResponseWriter, req *http.Request, username string, userinfo *auth.UserInfo) error {
	session := newAuthSession(h.authSessionLifetime, username, userinfo)
	httpSession, _ := h.authSessionStore.Get(req, authSessionKey)
	httpSession.Values["auth"] = session
	return httpSession.Save(req, w)
}

func (h *Server) withAuth(f func(http.ResponseWriter, *http.Request, *authSession)) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, req *http.Request) {
		httpSession, err := h.authSessionStore.Get(req, authSessionKey)
		if err != nil {
			http.Error(w, err.Error(), http.StatusInternalServerError)
			return
		}
		session, ok := httpSession.Values["auth"].(*authSession)
		if ok && session != nil && session.Valid() {
			f(w, req, session)
			return
		}
		httpSession.Options.MaxAge = -1
		_ = httpSession.Save(req, w)
		http.Redirect(w, req, makeLoginURL(req), http.StatusFound)
	})
}

func makeLoginURL(req *http.Request) string {
	// Just concatenate path and raw request string.
	v := make(url.Values)
	v.Set("r", req.URL.Path+"?"+req.URL.RawQuery)
	return "/login?" + v.Encode()
}

// Homepage handler. Authorizes an authenticated user to a service by
// signing a token with the user's identity. The client is redirected
// back to the service, with the signed token.
func (h *Server) handleHomepage(w http.ResponseWriter, req *http.Request, session *authSession) {
	// Extract the authorization request parameters from the HTTP
	// request.
	username := session.Username
	service := req.FormValue("s")
	destination := req.FormValue("d")
	nonce := req.FormValue("n")
	var groups, reqGroups []string
	if gstr := req.FormValue("g"); gstr != "" {
		reqGroups = strings.Split(gstr, ",")
		if len(reqGroups) > 0 && session.UserInfo != nil {
			groups = intersectGroups(reqGroups, session.UserInfo.Groups)
			// We only make this check here as a convenience to
			// the user (we may be able to show a nicer UI): the
			// actual group ACL must be applied on the destination
			// service, because the 'g' parameter is untrusted at
			// this stage.
			if len(groups) == 0 {
				http.Error(w, "Unauthorized", http.StatusUnauthorized)
				return
			}
		}
	}

	// Make the authorization request.
	token, err := h.loginService.Authorize(username, service, destination, nonce, groups)
	if err != nil {
		log.Printf("auth error: %v: user=%s service=%s destination=%s nonce=%s groups=%s", err, username, service, destination, nonce, req.FormValue("g"))
		http.Error(w, err.Error(), http.StatusBadRequest)
		return
	}

	session.AddService(service)
	_ = sessions.Save(req, w)

	// Redirect to service callback.
	callbackURL := serviceLoginCallback(service, destination, token)
	http.Redirect(w, req, callbackURL, http.StatusFound)
}

type logoutServiceInfo struct {
	URL  string
	Name string
}

func (h *Server) handleLogout(w http.ResponseWriter, req *http.Request, session *authSession) {
	var svcs []logoutServiceInfo
	for _, svc := range session.Services {
		svcs = append(svcs, logoutServiceInfo{
			Name: svc,
			URL:  serviceLogoutCallback(svc),
		})
	}

	data := map[string]interface{}{
		"CSRFField": csrf.TemplateField(req),
		"Services":  svcs,
		"IsPOST":    false,
	}
	if req.Method == "POST" {
		data["IsPOST"] = true

		// Clear the local session.
		httpSession, _ := h.authSessionStore.Get(req, authSessionKey)
		httpSession.Options.MaxAge = -1
		_ = httpSession.Save(req, w)
	}

	h.tpl.ExecuteTemplate(w, "logout.html", data)
}

func (h *Server) handleExchange(w http.ResponseWriter, req *http.Request) {
	curToken := req.FormValue("cur_tkt")
	curService := req.FormValue("cur_svc")
	curNonce := req.FormValue("cur_nonce")
	newService := req.FormValue("new_svc")
	newNonce := req.FormValue("new_nonce")
	reqGroups := strings.Split(req.FormValue("new_groups"), ",")

	token, err := h.loginService.Exchange(curToken, curService, curNonce, newService, newNonce, reqGroups)
	if err != nil {
		log.Printf("exchange error: %v", err)
		http.Error(w, err.Error(), http.StatusBadRequest)
		return
	}

	w.Header().Set("Content-Type", "text/plain")
	_, _ = io.WriteString(w, token)
}

// Handler returns the http.Handler for the SSO server application.
func (h *Server) Handler() http.Handler {
	m := mux.NewRouter()

	var lih, loh http.Handler
	lih = h.loginHandler
	loh = h.withAuth(h.handleLogout)
	if h.csrfSecret != nil {
		csrfW := csrf.Protect(h.csrfSecret)
		lih = csrfW(lih)
		loh = csrfW(loh)
	}
	m.Handle("/login", withDynamicHeaders(lih))
	m.Handle("/logout", withDynamicHeaders(loh))

	m.PathPrefix("/static/").Handler(http.StripPrefix("/static/", http.FileServer(&assetfs.AssetFS{
		Asset:     Asset,
		AssetDir:  AssetDir,
		AssetInfo: AssetInfo,
		Prefix:    "static",
	})))

	m.Handle("/exchange", withDynamicHeaders(http.HandlerFunc(h.handleExchange)))
	m.Handle("/", withDynamicHeaders(h.withAuth(h.handleHomepage)))

	return m
}

// A relatively strict CSP.
const contentSecurityPolicy = "default-src 'none'; img-src 'self' data:; script-src 'self'; style-src 'self'; connect-src 'self';"

func withDynamicHeaders(h http.Handler) http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		w.Header().Set("Pragma", "no-cache")
		w.Header().Set("Cache-Control", "no-store")
		w.Header().Set("Expires", "-1")
		w.Header().Set("Content-Security-Policy", contentSecurityPolicy)
		w.Header().Set("X-Frame-Options", "NONE")
		w.Header().Set("X-XSS-Protection", "1; mode=block")
		w.Header().Set("X-Content-Type-Options", "nosniff")
		h.ServeHTTP(w, r)
	})
}

// Parse the templates that are embedded with the binary (in bindata.go).
func parseEmbeddedTemplates() *template.Template {
	root := template.New("").Funcs(template.FuncMap{
		"json": toJSON,
	})
	files, err := AssetDir("templates")
	if err != nil {
		log.Fatalf("no asset dir for templates: %v", err)
	}
	for _, f := range files {
		b, err := Asset("templates/" + f)
		if err != nil {
			log.Fatalf("could not read embedded template %s: %v", f, err)
		}
		if _, err := root.New(f).Parse(string(b)); err != nil {
			log.Fatalf("error parsing template %s: %v", f, err)
		}
	}
	return root
}
